/* disk.c -- disk related routines
 *  $Header: /home/shenoy/work/research/disk-scheduling/simulator/RCS/disk.c,v 1.10 1997/07/12 20:10:39 shenoy Exp shenoy $
 * Last checkin by $Author: shenoy $
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Tue Apr  2 20:48:05 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Fri Jul 25 20:47:56 1997
 * Update Count    : 333
 * Status          : Unknown, Use with caution!
 * Notes           : Compile with a C++ compiler since some C++ features 
 *                   are used.
 */


#include "defns.h"
#include "disk.h"
#include "proto.h"

DiskArray diskArray;

void
initArray(struct arguments args, DiskArray & diskArray)
{
  int i;
 diskArray.numDisks  = args.numDisks;

 diskArray.disk = (Disk *) calloc(args.numDisks, sizeof(Disk) );
  if( null_ptr(diskArray.disk))
    error("Malloc failed in function initArray",E_MEM);


 diskArray.RAIDLevel = args.RAIDLevel;
 if(args.fixedBlockPlacement)
   diskArray.blockPlacement = FIXED;
 else
   diskArray.blockPlacement = VARIABLE;

 diskArray.placementPolicy = args.placementPolicy;
 diskArray.blockSize = args.blockSize;
 diskArray.parityGroupSize = args.parityGroupSize;
 diskArray.parityBlockSize = args.parityBlockSize;
 diskArray.stripeDegree = args.stripeDegree;

 diskArray.scheduleParam.schedulePolicy = args.schedulePolicy;
 diskArray.scheduleParam.diskPolicy  = args.diskPolicy;
 diskArray.scheduleParam.usePartitioning   = args.usePartitioning;
 diskArray.scheduleParam.partitionTime   = args.partitionTime;
 diskArray.scheduleParam.periodicSlice = args.periodicSlice;
 diskArray.scheduleParam.aperiodicSlice = args.aperiodicSlice;
 diskArray.scheduleParam.bestEffortSlice = args.bestEffortSlice;
 diskArray.scheduleParam.round = args.round;

 diskArray.diskParam  = args.diskParam;
 initDiskParams(diskArray.diskParam);

  /* _______ initialize tables __________*/
  generateParityGroupTable(args,diskArray);
  generateDataBlockTable(args, diskArray);
}


void
initDisk(Disk &disk)
{

  /* disk is initially UP */
  disk.status = UP;

  event_set(disk.notEmpty, "queue not empty", 3);
  event_set(disk.notFull, "queue not full", 3);
  disk.scheduledQNotEmpty = event("scheduled queue not empty");
  disk.scheduledQNotFull = event("scheduled queue not full");
  disk.workConserve = event("work conserving mode");

  disk.periodicQueue.head = disk.periodicQueue.tail = 0 ;
  disk.periodicQueue.lock = facility("periodic queue");
  disk.periodicQueue.notEmpty = &disk.notEmpty[0];
  disk.periodicQueue.notFull = &disk.notFull[0];
  
  disk.aperiodicQueue.head = disk.aperiodicQueue.tail = 0 ;
  disk.aperiodicQueue.lock = facility("aperiodic queue");
  disk.aperiodicQueue.notEmpty = &disk.notEmpty[1];
  disk.aperiodicQueue.notFull = &disk.notFull[1];

  disk.bestEffortQueue.head = disk.bestEffortQueue.tail = 0 ;
  disk.bestEffortQueue.lock = facility("best effort queue");
  disk.bestEffortQueue.notEmpty = &disk.notEmpty[2];
  disk.bestEffortQueue.notFull = &disk.notFull[2];

  disk.scheduledQueue.head = disk.scheduledQueue.tail = 0 ;
  disk.scheduledQueue.lock = facility("scheduled queue");
  disk.scheduledQueue.notEmpty = &disk.scheduledQNotEmpty;
  disk.scheduledQueue.notFull = &disk.scheduledQNotFull;
  disk.newRoundFlag=0;
}


void
generateParityGroupTable(struct arguments & args, DiskArray & diskArray)
{
  int row,col;

 diskArray.pgt = (ParityGroupTable **) malloc( sizeof(ParityGroupTable * ) * diskArray.parityGroupSize);
 if ( null_ptr(diskArray.pgt) ) 
   error("Malloc failed in fucntion: generateParityGroupTable", E_MEM);

 for(row = 0 ; row < diskArray.parityGroupSize ; row++){
   diskArray.pgt[row] = (ParityGroupTable *) malloc( sizeof(ParityGroupTable) * diskArray.numDisks);
   if ( null_ptr(diskArray.pgt[row] ))
     error("Malloc failed in fucntion: generateParityGroupTable", E_MEM);
 }
 for(col = 0; col < diskArray.numDisks ; col++){
   for(row = 0 ; row < diskArray.parityGroupSize; row++){
     if( row == (col% diskArray.parityGroupSize) && diskArray.RAIDLevel == RAID5)
       diskArray.pgt[ row ][ col ].type =  PARITY;
     else
       diskArray.pgt[ row ][ col ].type =  DATA;
   }
 }
}

void
generateDataBlockTable(struct arguments & args, DiskArray & diskArray)
{
 int i,row,col;

 diskArray.dbt = (DataBlockTable **) malloc( sizeof(DataBlockTable * ) * diskArray.parityGroupSize);
 if ( null_ptr(diskArray.dbt) ) 
   error("Malloc failed in fucntion: generateParityGroupTable", E_MEM);

 for(row = 0 ; row < diskArray.parityGroupSize; row++){
   diskArray.dbt[row] = (DataBlockTable *) malloc( sizeof(DataBlockTable) * diskArray.numDisks);
   if ( null_ptr(diskArray.dbt[row] ))
     error("Malloc failed in fucntion: generateParityGroupTable", E_MEM);
 }
 for(col = 0; col < diskArray.numDisks ; col++){
   i=0;
   for(row = 0 ; row < diskArray.parityGroupSize; row++){
     if( diskArray.pgt[ row][ col ].type == DATA ){
       diskArray.dbt[i][col] = row ;
       diskArray.pgt[row][col].dbt = i;
       i++;
     }
   }
 }
}



void
insertQueue(RequestQueue &queue, Request &request)
{
  /* calling funtion should check if queue is full before invoking this fn */
  if(isQueueFull(queue))
    error("Fatal error : attempt to insert into a full queue", E_GENERIC);


  if(isQueueEmpty(queue))
    set(*queue.notEmpty); /* this points to an entry in the event array */
   

  queue.queue[queue.tail] = request;
  queue.tail = (queue.tail+1)%MAX_REQUESTS; /* circular queue */
}


/* 
 * top level scheduler
 *  this process moves the requests from the top-level queues 
 * to the scheduled queues for a disk
 */
void 
SchedulerProcess(int diskNum)
{
  extern EVENT startSimulation;
  UsedSlices used;
  int periodicSched,aperiodicSched,bestEffortSched;
  double periodicSlice,aperiodicSlice,bestEffortSlice;
  int periodicQueueEmpty,aperiodicQueueEmpty,bestEffortQueueEmpty;
  Disk &disk=diskArray.disk[diskNum]; /* alias for the disk */

  /* ____ create the scheduler process ____*/
  create("scheduler");

  /* wait(startSimulation); */
  SchedulerMode(diskNum,NORMAL);
  computeSlices(periodicSlice,aperiodicSlice,bestEffortSlice);
  used.periodic = used.aperiodic = used.bestEffort      = 0.0;


  while(1){
    periodicSched = aperiodicSched = bestEffortSched = 0;
    resetSlices(diskNum,used); /* if new round, reset used to 0 */

    reserve(disk.periodicQueue.lock);
    if((periodicQueueEmpty = isQueueEmpty(disk.periodicQueue)) == 0)
      periodicSched = scheduleQueue(disk.periodicQueue,disk.scheduledQueue,periodicSlice,
			           used,disk.scheduleStats.idle);
    release(disk.periodicQueue.lock);

    reserve(disk.aperiodicQueue.lock);
    if((aperiodicQueueEmpty= isQueueEmpty(disk.aperiodicQueue)) == 0)
      aperiodicSched = scheduleQueue(disk.aperiodicQueue,disk.scheduledQueue,aperiodicSlice,
		                               used,disk.scheduleStats.idle);
    release(disk.aperiodicQueue.lock);
    
    reserve(disk.bestEffortQueue.lock);
    if((bestEffortQueueEmpty = isQueueEmpty(disk.bestEffortQueue)) == 0)
      bestEffortSched = scheduleQueue(disk.bestEffortQueue, disk.scheduledQueue,bestEffortSlice,
			              used,disk.scheduleStats.idle);
    release(disk.bestEffortQueue.lock); 


    /* sleep if all queues are empty, wait till the next round if slices are too small to schedule req*/
    if(bestEffortQueueEmpty && periodicQueueEmpty && aperiodicQueueEmpty ) 
      wait_any(disk.notEmpty); /* array of events */
    else if(periodicSched == 0 && aperiodicSched == 0 && bestEffortSched == 0) {
      SchedulerMode(diskNum,IDLE); /* switch to work conserving mode */
      hold(timeLeftInRound()+0.1); /* sleep till end of round */
      SchedulerMode(diskNum,NORMAL); /* normal mode */
    }
  }
}



void
WorkConservingSchedulerProcess(int diskNum)
{
  int insertedARequest;
  Request request;
  Disk &disk= diskArray.disk[diskNum];

  create("work conserving scheduler");

  while(1){
    wait(disk.workConserve);
    if(disk.scheduleStats.scheduler == IDLE && disk.scheduleStats.disk == IDLE){
      insertedARequest=0;
      while(!isQueueEmpty(disk.bestEffortQueue) && !insertedARequest){
	request = nextQueue(disk.bestEffortQueue);
	request.flags |= WORK_CONSERVE;
	if(!searchQueue(disk.scheduledQueue,request)){
	   insertCircularQueue(disk.scheduledQueue,request,disk.scheduledQueue.tail);
	   insertedARequest=1;
	}
	deleteQueue(disk.bestEffortQueue);
      }

      while(!isQueueEmpty(disk.aperiodicQueue) && !insertedARequest){
	request = nextQueue(disk.aperiodicQueue);
	request.flags |= WORK_CONSERVE;
	if(!searchQueue(disk.scheduledQueue,request)){
	  insertCircularQueue(disk.scheduledQueue,request,disk.scheduledQueue.tail);
	  insertedARequest=1;
	}
	deleteQueue(disk.aperiodicQueue);
      }

      while(!isQueueEmpty(disk.periodicQueue) && !insertedARequest){
	request = nextQueue(disk.periodicQueue);
	request.flags |= WORK_CONSERVE;
	if(!searchQueue(disk.scheduledQueue,request)){
	   insertCircularQueue(disk.scheduledQueue,request,disk.scheduledQueue.tail);
	   insertedARequest=1;
	}
	deleteQueue(disk.periodicQueue);
      }

    }      
  }
}


void
resetSlices(int diskNum,UsedSlices &used)
{
  Disk &disk=diskArray.disk[diskNum];
  if(disk.newRoundFlag){
    used.periodic = used.aperiodic = used.bestEffort = 0;
    disk.scheduleStats.periodicData = disk.scheduleStats.aperiodicData = 0;
    disk.scheduleStats.bestEffortData = disk.scheduleStats.data = 0;
    disk.scheduleStats.latency = disk.scheduleStats.idle = 0;
    disk.newRoundFlag = 0;
  }
}

void
computeSlices(double &periodicSlice, double &aperiodicSlice, 
	      double &bestEffortSlice)
{
  int round;
  double sum;

  if(!diskArray.scheduleParam.usePartitioning){ /* no partitioning => best effort scheduling*/
    periodicSlice = aperiodicSlice = bestEffortSlice = INFINITY;
    return;
  }

    round = diskArray.scheduleParam.round;
    sum = diskArray.scheduleParam.periodicSlice + 
               diskArray.scheduleParam.aperiodicSlice + 
               diskArray.scheduleParam.bestEffortSlice;

    periodicSlice    = diskArray.scheduleParam.periodicSlice    * round / sum;
    aperiodicSlice  = diskArray.scheduleParam.aperiodicSlice  * round / sum;
    bestEffortSlice = diskArray.scheduleParam.bestEffortSlice * round / sum;

}

int
scheduleQueue(RequestQueue &pendingQueue, RequestQueue &scheduledQueue, 
	           double slice, UsedSlices &used, double idle)
{
  Request request; 
  int  scheduled =0,success=1;  

  while(!isQueueEmpty(pendingQueue) &&  success ){
    request = nextQueue(pendingQueue); /* get req, but don't delete */
    if(!searchQueue(scheduledQueue,request))
      success = insertIntoScheduledQueue(scheduledQueue,request,slice,used,idle); /* insert */
    if(success){
      deleteQueue(pendingQueue); /* delete request */
        scheduled++;
    }
  }
  return scheduled;
}

int
insertIntoScheduledQueue(RequestQueue &queue, Request &request, double slice,  
		             UsedSlices &usedSlices,double idle)
{
     Estimates estimates;
     double estimate,additional,used,totalUsed,totalIdle;
     int retval=0;
     int round,pos;
 
     /* if requested block is also requested by another client , then do nothing */
     if(searchQueue(queue,request))
       return retval;

     /* compute some useful variables */
     round = diskArray.scheduleParam.round;
     totalIdle   = idle ; /* total time the disk was idle this round */
     idle = idle * slice / round; /* fraction of idle time charged to this class */
     totalUsed = usedSlices.periodic + usedSlices.aperiodic + usedSlices.bestEffort; 
     if(request.type == VIDEO)
       used = request.mode == PUSH ? usedSlices.periodic : usedSlices.aperiodic;
     else 
       used = usedSlices.bestEffort;
     
     reserve(queue.lock); /* acquire lock */   
     while(isQueueFull(queue)){
       release(queue.lock); /* release lock */
       wait(*queue.notFull); /* sleep */
       reserve(queue.lock); /* (re)acquire lock */   
     }
     /* at this point queue is guaranteed to be non-empty and locked */

     /* find the position to insert this request in the queue */
     switch(diskArray.scheduleParam.schedulePolicy){
     case LOOK:
     case CLOOK:
       pos = findScanPosition(queue,request);  
       break;
     case PLOOK: /* CLOOK with priority to text requests */
       pos = findPriorityScanPosition(queue,request); /* do priority SCAN for text */
       break;
     case PLOOKEDF: 
       pos = findPlookEdfPosition(queue,request);
       break;
     case FCFS:
       pos = findFcfsPosition(queue,request);
       break;
     }

     /* estimate  new service times if req is inserted */
     estimates = estimateServiceTime(queue,request,pos);
     additional = estimates.newReq + estimates.nextReq - estimates.old; 
     estimate    = diskArray.scheduleParam.partitionTime  ? estimates.newReq : 
                            estimateServiceTimeByte(request,estimates.newReq) ;

     /* insert if sufficient disk bw is available */
     if(!diskArray.scheduleParam.usePartitioning || 
	(totalUsed+totalIdle+additional <= round  && used+idle+estimate<=slice)){
                   insertCircularQueue(queue,request,pos);
                   updateServiceEstimate(queue,pos); /* update estimate of req AFTER pos */
	 updateUsedSlices(queue,pos,usedSlices,estimates);
	 retval=1;
       }

     release(queue.lock);
     return retval;
}


void
DiskProcess(int disk, Statistics &stats)
{
  double seek,xfr,rot,time,holdTime;
  Request req;
  DiskLocation loc = {0,0,0,0,0}; /* lbn,c,h,s,scan dirn (current disk state)*/

  /*___ create the disk process ____*/
  create("disk"); 
  
  DiskMode(disk,NORMAL);
  /* start high level scheduler */
  SchedulerProcess(disk);
  WorkConservingSchedulerProcess(disk);

  while(1){
    reserve(diskArray.disk[disk].scheduledQueue.lock);
    while(isQueueEmpty(diskArray.disk[disk].scheduledQueue)){

      release(diskArray.disk[disk].scheduledQueue.lock);
      DiskMode(disk,IDLE); /* queue is empty: switch to idle mode */
      time=clock;     /* time at which disk became idle */
      loc.sector = -1; 
      wait(*diskArray.disk[disk].scheduledQueue.notEmpty); /* wait 'til some req arrives */
      diskArray.disk[disk].scheduleStats.idle=min(clock-time,  timeElapsedInRound());
      DiskMode(disk,NORMAL);
      reserve(diskArray.disk[disk].scheduledQueue.lock);
    }
    /* at this point, there is atleast one req in the queue */
    if(isQueueFull(diskArray.disk[disk].scheduledQueue))
      set(*(diskArray.disk[disk].scheduledQueue.notFull)); 
    
    /* get a request from the schedule queue depending on the schedule policy */
    switch(diskArray.scheduleParam.diskPolicy){
    case FCFS: 
    case PLOOKEDF:
    case PLOOK: /* CLOOK with priority to text, queue is already sorted in plook order */
      req = deleteQueue(diskArray.disk[disk].scheduledQueue);
      break;
    case LOOK:
    case CLOOK:
    /* next req in SCAN/LOOK order*/
    req = getNextRequest(loc.block,loc.direction,
			 diskArray.disk[disk].scheduledQueue);
    if(req.blockNo == -1){
      /* no blocks in this direction */
      loc.direction = 1 - loc.direction; /* reverse direction */
      req = getNextRequest(loc.block,loc.direction,
			   diskArray.disk[disk].scheduledQueue); 
      if(diskArray.scheduleParam.diskPolicy == CLOOK) 
	loc.direction = 1 - loc.direction;
      if(req.blockNo == -1) 
	error("ScheduledQueue error: must have atleast one req in queue",E_GENERIC);
    }
    break;
    }

    release(diskArray.disk[disk].scheduledQueue.lock);
    
    /* process the request */
    seek = getSeek( loc.cylinder, getCylinder(req.blockNo) );
    rot  = getRot(loc.sector, getFirstSector(req.blockNo),seek);
    xfr  = getXfr(req.blockSize);
    holdTime = seek + rot + xfr;
    /* compute new position of the disk head */
    loc.block = getLastBaseBlock(req.blockNo,req.blockSize);/* last base blk */
    loc.cylinder = getCylinder(loc.block);
    loc.track    = getTrack(loc.block);
    loc.sector   = getLastSector(loc.block, diskArray.blockSize); /* head is at the end of this sector  */
    loc.sector  = (loc.sector+1)% diskArray.diskParam.sectorsTrack; /* hd is at the start of this sector*/
    
    /* if request is serviced in workconserving mode: update idle time (since disk is officially idle)*/
    if(req.flags & WORK_CONSERVE)
      diskArray.disk[disk].scheduleStats.idle += holdTime;


#ifdef DEBUG
    printf("time = %f, blk = %d disk %d : holding for %f [%f + %f + %f]  blksize = %d \n",clock,req.blockNo,disk,holdTime,seek,rot,xfr,req.blockSize);
#endif
    
    hold(holdTime);
    updateDiskStatistics(disk,req,holdTime,stats);
    updateClientStatistics(req,holdTime,stats);
  }

}


int
isQueueEmpty(RequestQueue &queue)
{
  return (queue.head == queue.tail);
}

int
isQueueFull(RequestQueue &queue)
{
  return ((queue.tail+1)%MAX_REQUESTS == queue.head);
}


Request &nextQueue( RequestQueue &queue)
{
  return queue.queue[queue.head];
}

Request &deleteQueue(RequestQueue &queue)
{

  if(isQueueFull(queue))
    set(*queue.notFull); 
  
   queue.head = (queue.head+1)%MAX_REQUESTS;
   return queue.queue[(MAX_REQUESTS+queue.head-1)%MAX_REQUESTS];
}


/* search queue if block exists in the queue */
int
searchQueue(RequestQueue &queue, Request &request)
{
   register int i ;
   /* NOTE : currently this  functionality has been disabled; define it carefully for raid5 */
   return 0;
   i = queue.head;
   while( i != queue.tail )
     if(request.blockNo == queue.queue[i].blockNo && request.blockSize == queue.queue[i].blockSize)
       return 1;
     else
       i = (i+1) % MAX_REQUESTS;
   return 0;
}


int
getLastBaseBlock(int blk,int size)
{
  return blk + size/diskArray.blockSize - 1;
}

int
getCylinder(int blk)
{
  return (blk* (diskArray.blockSize * 1024 / diskArray.diskParam.bytesSector)) / 
            (diskArray.diskParam.sectorsTrack * diskArray.diskParam.tracksCyl);
}

int
getTrack(int blk)
{
  int trk = (blk* (diskArray.blockSize * 1024 / diskArray.diskParam.bytesSector)) % 
                 (diskArray.diskParam.sectorsTrack *  diskArray.diskParam.tracksCyl);
  return trk / diskArray.diskParam.sectorsTrack;
}

int
getSector(int blk)
{
  int sector = (blk* (diskArray.blockSize * 1024 / diskArray.diskParam.bytesSector)) % 
                       (diskArray.diskParam.sectorsTrack *  diskArray.diskParam.tracksCyl);
  return sector % diskArray.diskParam.sectorsTrack;
}

int 
getFirstSector(int blk)
{
  return getSector(blk);
}

int
getLastSector(int blk,int size)
{
  return getSector(blk) + (size *1024 / diskArray.diskParam.bytesSector) - 1 ;
}

int
isParity(int disk, int block)
{
  return (diskArray.pgt[block % diskArray.parityGroupSize][disk].type == PARITY);
}

int
isData(int disk, int block)
{
  return (diskArray.pgt[block % diskArray.parityGroupSize][disk].type == DATA);
}

void
initDiskParams(DiskParam &diskParam)
{
  diskParam.A = (-10.0 * diskParam.minSeek + 15 * diskParam.avgSeek 
		 - 5 * diskParam.maxSeek)/(3.0 * sqrt(diskParam.numCyl));
  diskParam.B = (7.0 * diskParam.minSeek - 15 * diskParam.avgSeek 
		 + 8 * diskParam.maxSeek)/(3.0 * diskParam.numCyl);
  diskParam.C = diskParam.minSeek;
  diskParam.diskSize = (diskParam.bytesSector * diskParam.sectorsTrack *  diskParam.tracksCyl ) / 1024;
  diskParam.diskSize *= diskParam.numCyl; /* do this here to prevent integer overflow on prev line */
  diskParam.diskSize -= 1024; /* set aside some disk space for stroing super blk etc */
} 

double
getSeek(int a, int b)
{
  int x;
  x = a < b ? b-a : a-b;

  if ( x == 0 )
    return 0.0;
  if (x == 1)
    return diskArray.diskParam.C;
  return ((diskArray.diskParam.A * sqrt((double)x - 1.0))  +
	  (diskArray.diskParam.B *((double)x - 1.0)) + diskArray.diskParam.C);
}

/* estimate rot latency given that disk starts at sector a, seeks for 
 * time = seek, and must reach sector b
 */
double
getRot(int a, int b, double seek)
{
  double time;
  int sec;
  /* if a == -1 then exact location of disk is not known; compute it */
  if(a==-1){
    time = clock - floor(clock/diskArray.diskParam.rotLatency)*diskArray.diskParam.rotLatency;
    a = (int) (time * diskArray.diskParam.sectorsTrack / diskArray.diskParam.rotLatency);
  }

  /* compute sector at end of seek */
  sec = a;
  /* uncomment following line if disk does NOT have rotational skewing */
  /*sec += (int) (seek * diskArray.diskParam.sectorsTrack / (diskArray.diskParam.rotLatency));*/
  sec %= diskArray.diskParam.sectorsTrack;

  /* estimate rot latency */
  if(b >= sec)
    return (b-sec) * (diskArray.diskParam.rotLatency) / diskArray.diskParam.sectorsTrack;
  
  return (diskArray.diskParam.sectorsTrack - (sec-b)) * (diskArray.diskParam.rotLatency) / diskArray.diskParam.sectorsTrack;
}
  
double
getXfr(int size)
{
  /* size is in kB amd maxXfrRate in MB/s; return time in msec */
  return size * 1000.0 / (diskArray.diskParam.maxXfrRate * 1024);
}

/* LOOK scheduling : get request with LBN closest to block # block in
 * the specified scan direction
 * CLOOK: returns nearest blk in fwd direction; furthest block in reverse 
 * dirn 
*/
Request getNextRequest(int block,int direction, RequestQueue &queue)
{
  int i,min,max,diff,pos;
  Request req;
  min = INFINITY;
  max=-1;

  req.blockNo = pos = -1;
  i = queue.head;
  while(i != queue.tail){
    if(direction == 0 && queue.queue[i].blockNo >= block){
      /* return nearest block in forward dirn */
      diff = queue.queue[i].blockNo - block;
      if( diff < min ){
	min = diff;
	req  = queue.queue[i];
	pos = i;
      }
    }
    else if(direction == 1 && queue.queue[i].blockNo <= block){
      /* in the reverse dirn : return nearest blk if LOOK; furthest block if CLOOK */
      diff = block - queue.queue[i].blockNo;
      if(diskArray.scheduleParam.diskPolicy == LOOK){
	if(diff < min){
	  min = diff;
	  req  = queue.queue[i];
	  pos = i;
	}
      }
      else if(diskArray.scheduleParam.diskPolicy == CLOOK){
	if(diff > max){
	  max = diff;
	 req  = queue.queue[i];
	  pos = i;
	}
      } 
      }
    i = (i+1)%MAX_REQUESTS;
  }


  if(pos != -1){ /* delete req from queue */
    i = pos;
    while(i != queue.tail){
      queue.queue[i] = queue.queue[(i+1)%MAX_REQUESTS];
      i = (i+1)%MAX_REQUESTS;
    }
    queue.tail = (MAX_REQUESTS + queue.tail - 1) %MAX_REQUESTS;
  }

  return req;
}




/* find position to insert (SCAN order) */
int
findScanPosition(RequestQueue &queue, Request &req)
{
   int i;
   i = queue.head;
   while(i != queue.tail)
     if(queue.queue[i].blockNo < req.blockNo)
       i = (i+1)%MAX_REQUESTS;
     else
       break;
   return i;
}

/* find position to insert in priority SCAN order
   i.e., insert Text request at head of the queue in SCAN order */
int
findPriorityScanPosition(RequestQueue &queue, Request &req)
{
   int i;
   i = queue.head;

     /* if Video, advance i to first video request */
   if(req.type != TEXT)
     while(i != queue.tail && queue.queue[i].type != req.type)
       i = (i+1)%MAX_REQUESTS;

   while(i != queue.tail)
     if(queue.queue[i].blockNo < req.blockNo && queue.queue[i].type == req.type)
       i = (i+1)%MAX_REQUESTS;
     else
       break;
   return i;
}

int
findPlookEdfPosition(RequestQueue &queue, Request &req)
{
  int i,done,first,last;
  double time,additional;
  Estimates estimates;
  const int max=MAX_REQUESTS;
  
  if(isQueueEmpty(queue))
    return queue.tail;

  if(req.type == VIDEO){ /* insert in SCAN-EDF order */
    i = (max+queue.tail-1)%max; 
    /* scan backwards to find the last video request */
    while(i != queue.head && queue.queue[i].type != req.type )
      i = (max + i - 1) % max;

    /* if the req is video and has the same deadline : find the 1st req with same deadline */
    while(i != queue.head && queue.queue[i].type == req.type && 
                 queue.queue[i].deadline == req.deadline)
	i = (max + i - 1) % max;

    /* case 1: found  req with same deadline => insert in SCAN order */
    if(queue.queue[i].type == req.type && queue.queue[i].deadline == req.deadline){
      while(i != queue.tail && queue.queue[i].type == req.type && 
	    queue.queue[i].deadline == req.deadline && queue.queue[i].blockNo < req.blockNo)
	i = (i + 1) % max;
      return i;
    }
    /* case 2: no video req have the same deadline => insert at the last possible place without
        violating deadline */
    time= queue.queue[i].type == VIDEO ? 
               queue.queue[i].deadline-queue.queue[i].serviceEstimate : clock;
    i = queue.queue[i].type == VIDEO ? (i+1)%max : i;
    while(i != queue.tail && time+queue.queue[i].serviceEstimate+
	                                getServiceTime(queue.queue[i],req) < req.deadline){
      time += queue.queue[i].serviceEstimate;
      i = (i + 1) % max;
    }
    return i;
  }
  else{ /*request == TEXT */
    done=0;
    i = queue.head;
    time=clock;
    while(!done){
      first = i ;
      /* find the last text req in this seq */
      while( i != queue.tail && queue.queue[i].type == TEXT){
	time += queue.queue[i].serviceEstimate;
	i = (i+1)%max;
      }
      last = i;
      /* find position in scan order */
      while(first != last && queue.queue[i].blockNo < req.blockNo)
	first = (first+1)%max;
      estimates= estimateServiceTime(queue,req,first);
      additional = estimates.newReq +estimates.nextReq-estimates.old;
      /* can we insert it here (i.e., first) ? */
      /* first add servTime of all video req with same deadlines */
      while(i != queue.tail && queue.queue[i].type != TEXT && 
	queue.queue[i].deadline == queue.queue[last].deadline){
	time+=queue.queue[i].serviceEstimate;
	i = (i+1)%max;
      } 
      /* insert if last==tail  or there is slack */
      if(last == queue.tail ||  time+additional < queue.queue[last].deadline) 
	return first;

    } /* while(!done) */
  }
}

/* insert at the tail of the queue */
int
findFcfsPosition(RequestQueue &queue, Request &req)
{
    return queue.tail;
}


double
estimateServiceTimeByte(Request &request , double servTime)
{
  int data;
  double overhead;
  Disk &disk=diskArray.disk[request.disk];

  overhead = (disk.scheduleStats.latency + servTime - getXfr(request.blockSize)) / 
                       (disk.scheduleStats.data + request.blockSize) + getXfr(1);
  /*  data = request.type == TEXT ? disk.scheduleStats.bestEffortData :
    request.mode == PUSH ?disk.scheduleStats.periodicData : disk.scheduleStats.aperiodicData;*/
   return request.blockSize * (overhead+getXfr(1));
}

Estimates
estimateServiceTime(RequestQueue &queue, Request &req,int pos)
{

  int prev,next;
  double seek,rot,xfr;
  Request prevReq,nextReq;
  Estimates estimate;

  /* note : assume that the disk is scanning from prevReq to req to NextReq 
   * this will not be true if it is scanning in reverese direction, but estimate will
   * be approximately the same 
   */
  if(queue.head != pos){
    /* not the first request */
    prev = (MAX_REQUESTS + pos - 1) % MAX_REQUESTS;
    prevReq = queue.queue[prev];
    seek = getSeek(getCylinder(prevReq.blockNo),getCylinder(req.blockNo));
    rot    = getRot(getLastSector(prevReq.blockNo,prevReq.blockSize),
	            getFirstSector(req.blockNo),seek);
    xfr    = getXfr(req.blockSize);
  }
  else{
    /* first request */
    seek = diskArray.diskParam.avgSeek;
    rot    = diskArray.diskParam.rotLatency/2;
    xfr    = getXfr(req.blockSize);
  }

  req.serviceEstimate = seek + rot + xfr;
  estimate.newReq = seek + rot + xfr;

  estimate.old = estimate.nextReq = 0;
  if(queue.tail != pos){
    nextReq = queue.queue[ pos ];
    estimate.old = nextReq.serviceEstimate; /* old estimate for req at pos */
    seek = getSeek(getCylinder(req.blockNo),getCylinder(nextReq.blockNo));
    rot    = getRot(getLastSector(req.blockNo,req.blockSize),
                              getFirstSector(nextReq.blockNo),seek);
    xfr    = getXfr(nextReq.blockSize);
    estimate.nextReq = (seek + rot + xfr); /* new estimate */
  }

  return estimate;

}  

void
updateServiceEstimate(RequestQueue &queue, int pos)
{

 int next;
 Request nextReq,req;
 double seek,xfr,rot;

 next = (pos + 1 ) % MAX_REQUESTS;
 if(next != queue.tail){
   req = queue.queue[pos];
   nextReq = queue.queue[ next ];

    seek = getSeek(getCylinder(req.blockNo),getCylinder(nextReq.blockNo));
    rot    = getRot(getLastSector(req.blockNo,req.blockSize),
                              getFirstSector(nextReq.blockNo), seek);
    xfr    = getXfr(nextReq.blockSize);
    nextReq.serviceEstimate = seek + rot + xfr;
 }
}







/* NOTE: locking is not really needed since the process is guaranteed
 * not to be preempted
 */
int
insertIntoQueue(RequestQueue &queue, Request &request)
{  

  int success=0;
  /* if requested block is also requested by another client , then do nothing */
  if(searchQueue(queue,request))
    return success;

  reserve(queue.lock); /* acquire lock */   
  while(isQueueFull(queue)){
      release(queue.lock); /* release lock */
      wait(*queue.notFull); /* sleep */
      reserve(queue.lock); /* (re)acquire lock */   
    }

  /* at this point queue is guaranteed to be non-full and locked */
  insertQueue(queue,request);
  release(queue.lock);
  success=1;
  return success;
}


void
insertCircularQueue(RequestQueue &queue, Request &request, int pos)
{
  int i,tail,max = MAX_REQUESTS;
  
   /* calling funtion should check if queue is full before invoking this fn */
  if(isQueueFull(queue))
    error("Fatal error : attempt to insert into a full queue", E_GENERIC);
  if(isQueueEmpty(queue))
    set(*queue.notEmpty); /* this points to an entry in the event array */
   
  if(pos == queue.tail){
    queue.queue[queue.tail] = request;
    queue.tail = (queue.tail+1)%MAX_REQUESTS; /* circular queue */
 }
  else{
    /* insert in the middle of the queue : must shift elements @ [position,tail)
       right by 1 */
    tail = (max + queue.tail - 1)% max;
    for(i=tail; i !=  (max  +  pos - 1)%max; i = (max + i -1)% max )
      queue.queue[ (i+1)% max ] = queue.queue[i];
    /* ____ insert request */
    queue.queue[ pos ] = request;
    queue.tail = (queue.tail+1)%MAX_REQUESTS; /* circular queue */
  }

}


/* simlate a failed disk : disk fails in the specified round */
void
FailureProcess(struct arguments &args, DiskArray &diskArray)
{
  int disk;

  create("failure process");

  /* wait(startSimulation);*/

  hold(args.failureRound);
  if(args.failureRound < args.simulationLength){
    disk = random_int(0,diskArray.numDisks-1);
    diskArray.disk[disk].status = DOWN;
    hold(args.simulationLength - args.failureRound);
  }
}

void
RoundTimerProcess()
{
  int i;
  create("timer process");
  while(1){
    for(i=0;i<diskArray.numDisks;i++)
      diskArray.disk[i].newRoundFlag= 1;
    hold(diskArray.scheduleParam.round);
  }

}



void
updateUsedSlices(RequestQueue &queue, int pos, UsedSlices &usedSlices, Estimates estimates)
{

  if(diskArray.scheduleParam.partitionTime)
    updateUsedSlicesByTime(queue,pos,usedSlices,estimates);
  else 
    updateUsedSlicesByByte(queue,pos,usedSlices,estimates);
}

void
updateUsedSlicesByTime(RequestQueue &queue, int pos, UsedSlices &usedSlices, Estimates estimates)
{
  int next;
  double excess;

  if(queue.queue[pos].type == VIDEO){
    if(queue.queue[pos].mode == PUSH)
      usedSlices.periodic += estimates.newReq;
    else /* PULL */
      usedSlices.aperiodic += estimates.newReq;
  }
  else /* TEXT */
    usedSlices.bestEffort += estimates.newReq;
  
    /* return excess charged back to class containing nextReq (excess can be -ve)*/
  next = (pos +1)%MAX_REQUESTS;
  if(next != queue.tail){
    excess = estimates.old - estimates.nextReq;
    if(queue.queue[next].type == VIDEO){
      if(queue.queue[next].mode == PUSH)
	usedSlices.periodic -= excess;
      else /* PULL */
	usedSlices.aperiodic -= excess;
    }
    else /* TEXT */
      usedSlices.bestEffort -= excess;;
  }  
}

/* partition bytes */
void
updateUsedSlicesByByte(RequestQueue &queue, int pos, UsedSlices &usedSlices, Estimates estimates)
{
  double latencyPerKiloByte; 
  Disk &disk = diskArray.disk[queue.queue[pos].disk];

    disk.scheduleStats.latency += estimates.newReq  +estimates.nextReq - estimates.old;
    /* deduct xfr time to get latency (xfr of nextReq and old will cancel out -- ignore them) */
    disk.scheduleStats.latency -= getXfr(queue.queue[pos].blockSize); 
    disk.scheduleStats.data  += queue.queue[pos].blockSize; /* blockSize is in KB */
    if(queue.queue[pos].type == VIDEO){
      if(queue.queue[pos].mode == PUSH) 
	disk.scheduleStats.periodicData += queue.queue[pos].blockSize; 
      else
	disk.scheduleStats.aperiodicData += queue.queue[pos].blockSize; 
    }
    else
      disk.scheduleStats.bestEffortData += queue.queue[pos].blockSize; 

    latencyPerKiloByte = disk.scheduleStats.latency / disk.scheduleStats.data ;
    /* now update slices */
    usedSlices.periodic    = (latencyPerKiloByte+getXfr(1)) * disk.scheduleStats.periodicData;
    usedSlices.aperiodic  = (latencyPerKiloByte+getXfr(1)) * disk.scheduleStats.aperiodicData;
    usedSlices.bestEffort = (latencyPerKiloByte+getXfr(1)) * disk.scheduleStats.bestEffortData;
}



double timeElapsedInRound(void)
{
   return clock-((long)clock/diskArray.scheduleParam.round)*diskArray.scheduleParam.round;
  /*  return    ((int) (ceil(clock)+0.1)) % diskArray.scheduleParam.round ;*/
}

double timeLeftInRound(void)
{
  return diskArray.scheduleParam.round - 
    clock+((long)clock/diskArray.scheduleParam.round)*diskArray.scheduleParam.round;
}


void
SchedulerMode(int disk,int mode)
{
  if(!diskArray.scheduleParam.usePartitioning)
    return;
  if(mode == IDLE){
    diskArray.disk[disk].scheduleStats.scheduler=IDLE;
    set(diskArray.disk[disk].workConserve);
  }
  else{
    diskArray.disk[disk].scheduleStats.scheduler=NORMAL;
    clear(diskArray.disk[disk].workConserve);
  }
}

void
DiskMode(int disk,int mode)
{
  
  if(!diskArray.scheduleParam.usePartitioning)
    return;
  if(mode == IDLE){
    diskArray.disk[disk].scheduleStats.disk=IDLE;
    set(diskArray.disk[disk].workConserve);
  }
  else{
    diskArray.disk[disk].scheduleStats.disk=NORMAL;
    clear(diskArray.disk[disk].workConserve);
  }
}


double
getServiceTime(Request &req1,Request &req2)
{
  double seek,rot,xfr;
   seek=getSeek(getCylinder(req1.blockNo),getCylinder(req2.blockNo));
   rot= getRot(getLastSector(req1.blockNo,req2.blockSize),
	       getFirstSector(req2.blockNo), seek);
   xfr= getXfr(req2.blockSize);
   return seek+rot+xfr;
}
