/* inpgen.c -- interactively generate an input file for the simulator
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Sun May 12 15:12:05 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Mon Jun 23 15:03:00 1997
 * Update Count    : 39
 * Status          : Unknown, Use with caution!
 */

#include "defns.h"
#include "disk.h"
#include "proto.h"


int
main(int argc, char *argv[])
{
  struct arguments args;
  getSimulationParameters(args);

}  


void
getSimulationParameters(struct arguments &args)
{
  int c;
  int ireply;
  double dreply;
  char array[MAX_STR];
 FILE *fp;

  fp = fopen("input","w");
  if(fp == NULL){
    error("Unable to open input file for writing", E_IO);
  }
  else
    fprintf(stderr, "The simulation parameters will be written to file \"input'\"  \n");

  printf("Enter number of disks : ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf( "Enter RAID Level [0-5] : ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);
  switch(ireply){
  case 0: args.RAIDLevel  = RAID0; break;
  case 1: args.RAIDLevel  = RAID1; break;
  case 2: args.RAIDLevel  = RAID2; break;
  case 3: args.RAIDLevel  = RAID3; break;
  case 4: args.RAIDLevel  = RAID4; break;
  case 5: args.RAIDLevel  = RAID5; break;
  default: args.RAIDLevel  = RAID0; break;
  }


  printf("Enter top level scheduling policy [ LOOK = 0, CLOOK = 1, FCFS = 2, PLOOK=3 PLOOKEDF=4 ]: ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf("Enter disk scheduling policy [ LOOK = 0, CLOOK = 1, FCFS = 2 ]: ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf("Enter disk placement policy [round robin = 0, random = 1]: ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf( "Enter base block size (kB) : ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf( "Use fixed block size (y/n) ");
  scanf ( "%s",array);
  fprintf(fp,"%s\n",array);


  if( ( array[0] == 'y' ) || (array[0] == 'Y')){
    printf( "Enter CM block size (kB) ");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);
  }
  else{
    printf("Allow fragmentation in variable-size blocks ? (y/n) ");
    scanf ("%s", array);
    fprintf(fp,"%s\n",array);
  }

  if(args.RAIDLevel >= RAID1){
    printf("Enter parity group size ");
    scanf ("%d",&ireply);
    fprintf(fp,"%d\n",ireply);
  }

  printf( "Stripe across all disks ? (y/n) ");
  scanf ( "%s", array);
  fprintf(fp,"%s\n",array);
  if( (array[0] == 'y') || ( array[0] == 'Y'))
    args.stripeDegree = args.numDisks;
  else{
    printf( "Enter stripe degree ");
    scanf ("%d", &ireply);
    fprintf(fp,"%d\n",ireply);
  }

  printf("Use default disk (seagate elite 3 ) [y/n] :");
  scanf ("%s", array);
  fprintf(fp,"%s\n",array);
  if(! (array[0] == 'y' || array[0] == 'Y')){

    printf("Bytes per Sector (bytes) ");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);

    printf("Sectors per Track ");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);

    printf("Tracks per cylinder ");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);

    printf("Number of Cylinders ");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);

    printf( "Minimum seek (msec) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);

    printf( "Average seek (msec) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);

    printf( "Maximum seek (msec) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);


    printf( "Rotational Latency (msec) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);

    printf( "Max transfer rate (MB/s) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);


  }
  

  printf( "Enter round duration (msec) ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);
  printf( "Enter data rate (frames per sec) ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf( "Enter trace file name ");
  scanf ( "%s", array); 
  fprintf(fp,"%s\n", array);

  printf("Enter number of CM files on disk :");
  scanf("%d",&ireply);
  fprintf(fp,"%d\n",ireply);
  
  printf("Enter number of text files on disk :");
  scanf("%d",&ireply);
  fprintf(fp,"%d\n",ireply);
  
  printf("Enter size of text files on disk (kB):");
  scanf("%d",&ireply);
  fprintf(fp,"%d\n",ireply); 


  printf( "Assume all clients arrive at t = 0 (y/n) ");
  scanf ( "%s", array);
  fprintf(fp,"%s\n",array);

  args.emptySystem = (! (array[0] == 'y') || (array[0] == 'Y'));
  if( args.emptySystem){
    printf( "Enter interarrival time (msec) ");
    scanf ( "%lf", &dreply);
    fprintf(fp,"%f\n",dreply);
  }
  else{
    printf("Enter number of video clients :");
    scanf ( "%d", &ireply);
    fprintf(fp,"%d\n",ireply);

    printf( "Enter number of text clients :");
    scanf ( "%d", &ireply);
    args.numTextClients = ireply;
    fprintf(fp,"%d\n",ireply);

    if(args.numTextClients>0){
      printf( "Enter the avg request size of text requests (kB) :");
      scanf ( "%d", &ireply);
      fprintf(fp,"%d\n",ireply);
      
      printf( "Enter the avg duration between text requests (msec) ");
      scanf ( "%d", &ireply);
      fprintf(fp,"%d\n",ireply);
    }
  }

  printf( "Use Server Push for CM clients (y/n) ");
  scanf ( "%s",array);
  fprintf(fp,"%s\n",array);
  args.serverPush = (array[0] == 'y' || array[0] == 'Y');
  if(! args.serverPush ){ /* pull mode */
    printf("Enter the request size of client pull requests (kB) :");
    scanf("%d",&ireply);
    fprintf(fp,"%d\n",ireply);
  }

  printf( "Enter simulation length (# rounds) ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);

  printf( "Enter starting round  for collecting statistics ");
  scanf ( "%d", &ireply);
  fprintf(fp,"%d\n",ireply);
  

  printf("Enter seed ");
  scanf("%d",&ireply);
  fprintf(fp,"%d\n",ireply);

  if(args.RAIDLevel >= RAID1){
    printf("Simulate a disk failure ? [y/n] ");
    scanf( "%s",array);
    fprintf(fp,"%s\n",array);
    if((array[0] == 'y') || (array[0] == 'Y')){
      printf("Enter round in which a disk fails : ");
      scanf("%d",&ireply);
      fprintf(fp,"%d\n",ireply);
    }
  }

  printf( "Partition disk bandwidth (y/n) ");
  scanf ( "%s",array);
  fprintf(fp,"%s\n",array);
  
  if((array[0] == 'y') || (array[0] == 'Y')){
    printf( "Partition time (n => partition bytes) (y/n) ");
    scanf ("%s",array);
    fprintf(fp,"%s\n",array);
    printf("Enter periodic slice ");
    scanf("%lf",&dreply);
    fprintf(fp,"%f\n",dreply);
    
    printf("Enter aperiodic slice ");
    scanf("%lf",&dreply);
    fprintf(fp,"%f\n",dreply);
    
    printf("Enter best effort slice ");
    scanf("%lf",&dreply);
    fprintf(fp,"%f\n",dreply);
   
  
 }
  if(args.numTextClients > 0 || args.serverPush == 0){
    printf( "Produce verbose output (y/n) ");
    scanf ( "%s",array);
    fprintf(fp,"%s\n",array);
  }

  /* add any new options here ... */
  /* do sanity checks here */

}
