/* main.c -- grand unified simulator (top level routines)
 *  $Header: /home/shenoy/work/research/disk-scheduling/simulator/RCS/main.c,v 1.13 1997/07/12 20:13:45 shenoy Exp $
 *  Last check in by $Author: shenoy $ 
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Tue Apr  2 12:23:19 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Mon Jun 23 14:33:17 1997
 * Update Count    : 106
 * Status          : Unknown, Use with caution!
 * Notes           : Compile using a C++ compiler (since some C++ features are used)
 */


#include "defns.h"
#include "disk.h"
#include "proto.h"


char err_str[MAX_STR];
Statistics stats;

int
main(int argc, char *argv[])
{
  struct arguments args;
  
  getopts(argc, argv, args);
  initArray(args, diskArray);
  diskLayout(args, diskArray);
  simulate(args, diskArray,stats);  
  printReport(args,diskArray,stats);

}
 

void
getopts(int argc, char *argv[], struct arguments & args)
{

  extern char *optarg;
  extern int optind;
  int c;
  
  args.ifileName = (char *) NULL;
  while(( c = getopt(argc,argv,"f:")) != EOF ){
    switch(c){
    case 'f':
     args.ifileName = optarg;
     break;
   case '?':
     printHelp(argv[0]);
     error("Unknown argument",E_ARGS);
     break;
    }
  }

  /* open appropriate files for input and output */  
  openFiles(args);
  /* now read all simulation parameters */
  getSimulationParameters(args);  
}


void
printHelp(char *program)
{
  fprintf(stderr,"Usage : %s -f <simulation parameters file>\n",program);
}

void 
openFiles(struct arguments &args)
{

  /* if input file is specifed or input is redirected, 
     direct output to /dev/null */
  
  if(args.ifileName){
    args.ifp = fopen(args.ifileName,"r");
    if(args.ifp == NULL){
      sprintf(err_str,"error in opening input file : %s", args.ifileName);
      error(err_str,E_IO);
    }
    args.ofp = fopen("/dev/null","w");
  }
  else{
    if(isatty(0)){
      args.ifp = stdin;
      args.ofp = stdout;
    }
    else{
      args.ifp = stdin;
      args.ofp = fopen("/dev/null","w");
    }
  }

}

void
getSimulationParameters(struct arguments &args)
{
  char array[MAX_STR];

  fprintf(args.ofp,"Enter number of disks : ");
  fscanf (args.ifp, "%d", &args.numDisks);
  fprintf(args.ofp, "Enter RAID Level [0-5] : ");
  fscanf (args.ifp, "%d", &args.RAIDLevel);
  fprintf(args.ofp, "Enter top level scheduling policy: [ LOOK = 0, CLOOK = 1, FCFS = 2, PLOOK=3 PLOOKEDF=4 ]: ");
  fscanf(args.ifp,"%d",&args.schedulePolicy);
  fprintf(args.ofp, "Enter  disk scheduling policy: [ LOOK = 0, CLOOK = 1, FCFS = 2 ]: ");
  fscanf(args.ifp,"%d",&args.diskPolicy);

  fprintf(args.ofp,"Enter disk placement policy [round robin = 0, random = 1]: ");
  fscanf(args.ifp,"%d",&args.placementPolicy);

  fprintf(args.ofp, "Enter base block size (kB) : ");
  fscanf (args.ifp, "%d", &args.blockSize);
  args.parityBlockSize = args.blockSize;

  fprintf(args.ofp, "Use fixed block size (y/n) ");
  fscanf (args.ifp, "%s", array);
  args.fixedBlockPlacement = ( array[0] == 'y' ) || (array[0] == 'Y');
  if(args.fixedBlockPlacement){
    fprintf(args.ofp, "Enter CM block size (kB) ");
    fscanf (args.ifp, "%d", &args.CMblockSize);
    args.allowFragmentation = 0 ;
  }
  else{
    fprintf(args.ofp,"Allow fragmentation in variable-size blocks ? (y/n) ");
    fscanf (args.ifp, "%s", array);
    args.allowFragmentation = ( array[0] == 'y' ) || (array[0] == 'Y');
  }

  if(args.RAIDLevel >= RAID1){
    fprintf(args.ofp,"Enter parity group size ");
    fscanf (args.ifp,"%d", &args.parityGroupSize);
  }
  else
    args.parityGroupSize =  args.numDisks;
  
  fprintf(args.ofp, "Stripe across all disks ? (y/n) ");
  fscanf (args.ifp, "%s", array);
  if( (array[0] == 'y') || ( array[0] == 'Y'))
    args.stripeDegree = args.numDisks;
  else{
    fprintf(args.ofp, "Enter stripe degree ");
    fscanf (args.ifp, "%d", &args.stripeDegree);
  }

  fprintf(args.ofp,"Use default disk (seagate elite 3 ) [y/n] :");
  fscanf (args.ifp,"%s", array);
  if(! (array[0] == 'y' || array[0] == 'Y')){
    fprintf(args.ofp,"Bytes per Sector (bytes) ");
    fscanf (args.ifp, "%d", &args.diskParam.bytesSector);
    fprintf(args.ofp,"Sectors per Track ");
    fscanf (args.ifp, "%d", &args.diskParam.sectorsTrack);
    fprintf(args.ofp,"Tracks per cylinder ");
    fscanf (args.ifp, "%d", &args.diskParam.tracksCyl);
    fprintf(args.ofp,"Number of Cylinders ");
    fscanf (args.ifp, "%d", &args.diskParam.numCyl);
    fprintf(args.ofp, "Minimum seek (msec) ");
    fscanf (args.ifp, "%lf", &args.diskParam.minSeek);
    fprintf(args.ofp, "Average seek (msec) ");
    fscanf (args.ifp, "%lf", &args.diskParam.avgSeek);
    fprintf(args.ofp, "Maximum seek (msec) ");
    fscanf (args.ifp, "%lf", &args.diskParam.maxSeek);
    fprintf(args.ofp, "Rotational Latency (msec) ");
    fscanf (args.ifp, "%lf", &args.diskParam.rotLatency);
    fprintf(args.ofp, "Max transfer rate (MB/s) ");
    fscanf (args.ifp, "%lf", &args.diskParam.maxXfrRate);
  }
  else{ /* default disk (Seagate Elite 3) */
     args.diskParam.bytesSector  = SE3_bytesSector;
     args.diskParam.sectorsTrack = SE3_sectorsTrack;
     args.diskParam.tracksCyl    = SE3_tracksCyl ;
     args.diskParam.numCyl       = SE3_numCyl;
     args.diskParam.minSeek      = SE3_minSeek;
     args.diskParam.maxSeek      = SE3_maxSeek;
     args.diskParam.avgSeek      = SE3_avgSeek;
     args.diskParam.rotLatency   = SE3_rotLatency;
     args.diskParam.maxXfrRate   = SE3_maxXfrRate;
   }
  

  fprintf(args.ofp, "Enter round duration (msec) "); /* required even for client pull (in initStats()) */
  fscanf (args.ifp, "%d", &args.round);
  fprintf(args.ofp, "Enter data rate (frames per sec) ");
  fscanf (args.ifp, "%d", &args.dataRate);
  fprintf(args.ofp, "Enter trace file name ");
  fscanf (args.ifp, "%s", &args.traceFile); /* should be an array (not char *) */
 
  fprintf(args.ofp,"Enter number of CM files on disk :");
  fscanf(args.ifp,"%d",&args.numCMFiles);
  fprintf(args.ofp,"Enter number of text files on disk :");
  fscanf(args.ifp,"%d",&args.numTextFiles);
  args.numFiles  = args.numCMFiles + args.numTextFiles;
  fprintf(args.ofp,"Enter size of text files on disk (kB):");
  fscanf(args.ifp,"%d",&args.textFileSize);

  fprintf(args.ofp, "Assume all clients arrive at t = 0 (y/n) ");
  fscanf (args.ifp, "%s", array);
  args.emptySystem = (! (array[0] == 'y') || (array[0] == 'Y'));
  if( args.emptySystem){
    fprintf(args.ofp, "Enter interarrival time (msec) ");
    fscanf (args.ifp, "%lf", args.interArrival);
  }
  else{
    fprintf(args.ofp,"Enter number of video clients :");
    fscanf (args.ifp, "%d", &args.numCMClients);
    fprintf(args.ofp, "Enter number of text clients :");
    fscanf(args.ifp, "%d", &args.numTextClients);
    args.numClients = args.numCMClients + args.numTextClients;
    if(args.numTextClients>0){
      fprintf(args.ofp, "Enter the avg request size of text requests (kB) :");
      fscanf(args.ifp, "%d", &args.requestSize);
      fprintf(args.ofp, "Enter the avg duration between text requests (msec) ");
      fscanf(args.ifp, "%d", &args.holdTime);
    }
  }

  fprintf(args.ofp, "Use Server Push for CM clients (y/n) ");
  fscanf (args.ifp, "%s",array);
  args.serverPush = (array[0] == 'y') || (array[0] == 'Y');
  if(!args.serverPush){
    fprintf(args.ofp,"Enter the request size of client pull requests (kB) :");
    fscanf(args.ifp,"%d",&args.pullRequestSize);
  }
  fprintf(args.ofp, "Enter simulation length (# rounds) ");
  fscanf (args.ifp, "%d", &args.simulationLength);
  args.simulationLength *= args.round;
  fprintf(args.ofp, "Enter starting round  for collecting statistics ");
  fscanf (args.ifp, "%d", &args.statsStartTime);
  args.statsStartTime *= args.round;

  fprintf(args.ofp,"Enter seed ");
  fscanf(args.ifp,"%d",&args.seed);

  if(args.RAIDLevel >= RAID1){
    fprintf(args.ofp,"Simulate a disk failure ? [y/n] ");
    fscanf(args.ifp, "%s",array);
    if((array[0] == 'y') || (array[0] == 'Y')){
      fprintf(args.ofp,"Enter round in which a disk fails : ");
      fscanf(args.ifp,"%d",&args.failureRound);
    }
    else
      args.failureRound = args.simulationLength;

  }
  else{
    args.failureRound = args.simulationLength;
  }

  fprintf(args.ofp, "Partition disk bandwidth (y/n) ");
  fscanf (args.ifp, "%s",array);
  args.usePartitioning = (array[0] == 'y') || (array[0] == 'Y');

  if(args.usePartitioning){
    fprintf(args.ofp, "Partition time (n => partition bytes) (y/n) ");
    fscanf (args.ifp, "%s",array);
    args.partitionTime = (array[0] == 'y') || (array[0] == 'Y');

    fprintf(args.ofp,"Enter periodic slice ");
    fscanf(args.ifp,"%lf",&args.periodicSlice);
    
    fprintf(args.ofp,"Enter aperiodic slice ");
    fscanf(args.ifp,"%lf",&args.aperiodicSlice);
    
    fprintf(args.ofp,"Enter best effort slice ");
    fscanf(args.ifp,"%lf",&args.bestEffortSlice);

  }
    
    if(args.numTextClients > 0 || args.serverPush == 0){
      fprintf(args.ofp, "Produce verbose output (y/n) ");
      fscanf (args.ifp, "%s",array);
      args.verboseOutput = (array[0] == 'y') || (array[0] == 'Y');
    }
  /* add any new options here ... */

  /* do sanity checks here */

}











