/* proto.h -- function prototypes
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Mon May  6 19:29:12 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Tue Jun 24 23:19:38 1997
 * Update Count    : 60
 * Status          : Unknown, Use with caution!
 */


#ifndef _PROTO_H
#define _PROTO_H

#include "defns.h"
#include "disk.h"
#include "client.h"


/* ______ main.c _________ */
int main(int argc, char *argv[]);
void getopts(int argc, char *argv[], struct arguments &args);
void openFiles(struct arguments &args);
void getSimulationParameters(struct arguments &args);
void printHelp(char *program);


/* ______ disk.c _________ */
void initArray(struct arguments args, DiskArray & diskArray);
void initDisk(Disk &disk);
void generateParityGroupTable(struct arguments & args, DiskArray & diskArray);
void generateDataBlockTable(struct arguments & args, DiskArray & diskArray);
void insertQueue(RequestQueue &queue, Request &request);
void SchedulerProcess(int diskNum);
void WorkConservingSchedulerProcess(int diskNum);
void resetSlices(int diskNum, UsedSlices &used);
void computeSlices(double &periodicSlice, double &aperiodicSlice, double &bestEffortSlice);
int scheduleQueue(RequestQueue &pendingQueue, RequestQueue &scheduledQueue, 
		     double slice,UsedSlices &used,double idle);
int insertIntoScheduledQueue(RequestQueue &queue, Request &request,  double slice,
		                 UsedSlices &used,double idle);
void DiskProcess(int disk, Statistics &stats);
int  isQueueFull(RequestQueue &queue);
int  isQueueEmpty(RequestQueue &queue);
Request &nextQueue(RequestQueue &queue);
Request &deleteQueue(RequestQueue &queue);
int searchQueue(RequestQueue &queue, Request &request);
int  getLastBaseBlock(int blk, int size);
int  getTrack(int blk);
int  getSector(int blk);
int  getCylinder(int blk);
int  getLastSector(int blk, int size);
int  getFirstSector(int blk);
int isParity(int disk, int block);
int isData(int disk, int block);
void initDiskParams(DiskParam &diskParam);
double getSeek(int a, int b );
double getRot(int a, int b, double seek);
double getXfr(int size);
Request getNextRequest(int block,int direction, RequestQueue &queue);
int findScanPosition(RequestQueue &queue, Request &req);
int findPriorityScanPosition(RequestQueue &queue, Request &req);
int findPlookEdfPosition(RequestQueue &queue, Request &req);
int findFcfsPosition(RequestQueue &queue, Request &req);
double estimateServiceTimeByte(Request &request , double servTime);
Estimates estimateServiceTime(RequestQueue &queue, Request &req,int pos);
void updateServiceEstimate(RequestQueue &queue, int pos);
int insertIntoQueue(RequestQueue &queue, Request &request);
void insertCircularQueue(RequestQueue &queue, Request &request, int pos);
void FailureProcess(struct arguments &args, DiskArray &diskArray);
void RoundTimerProcess(void);
void updateUsedSlices(RequestQueue &queue, int pos, UsedSlices &usedSlices, 
		      Estimates estimates);
void updateUsedSlicesByTime(RequestQueue &queue, int pos, UsedSlices &usedSlices, 
			    Estimates estimates);
void updateUsedSlicesByByte(RequestQueue &queue, int pos, UsedSlices &usedSlices, 
			    Estimates estimates);
double timeElapsedInRound(void);
double timeLeftInRound(void);
void SchedulerMode(int disk,int mode);
void DiskMode(int disk,int mode);
double getServiceTime(Request &req1,Request &req2);

/* _______ layout.c _______ */
void diskLayout(struct arguments &args, DiskArray &diskArray);
void Layout(struct arguments &args, DiskArray &diskArray);
void readTraceFile(char *filename, FrameTableElem * frameTable, int &frameTableSize);
void initFrameTable(struct arguments &args);
void initFileTable(struct arguments & args);
void LayoutFiles(struct arguments &args,DiskArray &diskArray);
void FixedBlockLayout(struct arguments &args, DiskArray &diskArray);
void VariableBlockLayout(struct arguments &args, DiskArray &diskArray);
void deleteFile(int file, int *fileList, int &fileListSize);
void initLayoutVars(int **currentfp,int **currentbp, int **currentFrameResidue,
	       int **fileList, int **currentDisk, int **firstFrame, int **currentdb, int **permute);
void generatePermutation(int *array, int size);
void checkLastBlock(struct arguments &args,  DiskArray &diskArray);

/* _______ client.c ________ */
void simulate(struct arguments &args, DiskArray &diskArray, Statistics &stats);
void ClientProcess(int client);
double getHoldTime(Client &client);
void generateRequest(RequestList &requestList, Client &client);
void generateTextRequest(RequestList &requestList, Client &client);
void generateCMPushRequest(RequestList & requestList, Client &client);
void generateCMPullRequest(RequestList & requestList, Client &client);
void issueRequest(RequestList &requestList);
void initClient(struct arguments &args,Client &client, FileType type);
int nextMetaBlock(int block, int file);
void retrieveParityGroup(Request &request);
void coalesceRequests(RequestList &requestList);
/* _____________ stats.c ___________*/
void printReport(struct arguments &args, DiskArray &array, Statistics &stats);
void updateDiskStatistics( int disk, Request &req, double holdTime, Statistics &stats);
void updateClientStatistics(Request &req, double holdTime, Statistics &stats);
void StatisticsProcess(struct arguments &args, DiskArray &diskArray, Statistics &stats);
void collectStatistics(DiskArray &diskArray, Statistics &stats);
void initStats(struct arguments &args, DiskArray &diskArray, Statistics &stats);
void registerRequest(RequestList &req);
int findEmptyPendingRequestSlot(int client);
int findPendingRequest(Statistics &stats, int client, double clock);
void incrementPendingRequestSlot(Request &req);
void decrementPendingRequestSlot(Request &req);
#endif





