/* stats.c --  Collect and analyze simulation statistics
 * $Header: /home/shenoy/work/research/disk-scheduling/simulator/RCS/stats.c,v 1.6 1997/07/12 20:14:20 shenoy Exp $
 * Last checkin by $Author: shenoy $
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Sat May 18 11:28:34 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Tue Dec  3 21:02:34 1996
 * Update Count    : 114
 * Status          : Unknown, Use with caution!
 */


#include "defns.h"
#include "disk.h"
#include "proto.h"


void
updateDiskStatistics(int disk, Request &req, double holdTime, Statistics &stats)
{
  stats.disk[disk].busyTime += holdTime;
  switch(req.type){
  case VIDEO:
    stats.disk[disk].video += req.blockSize;
    break;
  case TEXT:
    stats.disk[disk].text += req.blockSize;
    break;
  }
}


void
updateClientStatistics(Request &req, double holdTime, Statistics &stats)
{
  int i;



    i = req.pendingRequestIndex;
    stats.pendingRequests[req.client][i]--;
    if(stats.pendingRequests[req.client][i]<0)
      error("updateClientStats():stats.pendingRequests[req.client][i]<0",E_GENERIC);
    if(stats.pendingRequests[req.client][i]==0){ /* all physical req have finished */
      switch(req.type){ /* per request statistics */
      case VIDEO:
	record(clock-req.arrivalTime,stats.videoResponseTime);
	break;
      case TEXT:
	record(clock-req.arrivalTime,stats.textResponseTime);
	break;
      }
    }
    
    if(req.type == TEXT){ /* per disk block statistics */
      record(clock-req.arrivalTime-holdTime,stats.textQueuingDelay);
      record(holdTime,stats.textServiceTime);
    }
}

void
StatisticsProcess(struct arguments &args, DiskArray &diskArray, Statistics &stats)
{
  extern EVENT startSimulation;

 create("statistics");

 /* wait(startSimulation);*/
 /* hold for certain number of rounds before collecting statistics */
 hold(args.statsStartTime);

 /* gather stats 0.1 msec before a new round begins */ 
 hold((double) (args.round - 0.1));
 while(1){
   collectStatistics(diskArray,stats);
   hold(args.round);
 }

}

 
void
collectStatistics(DiskArray &diskArray, Statistics &stats)
{
  int i;
  double maxBusyTime,minBusyTime,sumBusyTime;
  double maxBlocks,minBlocks,sumBlocks,maxVideoBlocks,minVideoBlocks,sumVideoBlocks;
  double maxTextBlocks, minTextBlocks, sumTextBlocks;
  
  maxBusyTime = maxBlocks = maxTextBlocks = maxVideoBlocks = 0;
  sumBusyTime  = sumBlocks =  sumTextBlocks = sumVideoBlocks = 0;
  minBusyTime  = minBlocks = minTextBlocks = sumTextBlocks = INFINITY;


  for(i=0;i<diskArray.numDisks;i++){
    /* per round statistics */
#ifdef STATS_DEBUG
 printf("disk = %d > busy = %f, blocks = %d\n",i,stats.disk[i].busyTime,(stats.disk[i].video + stats.disk[i].text)/diskArray.blockSize);
#endif
    maxBusyTime = max(stats.disk[i].busyTime, maxBusyTime);
    minBusyTime  = min(stats.disk[i].busyTime, minBusyTime);
    sumBusyTime  +=  stats.disk[i].busyTime;

    maxBlocks = max((stats.disk[i].video + stats.disk[i].text), maxBlocks);
    minBlocks = min((stats.disk[i].video + stats.disk[i].text), minBlocks);
    sumBlocks += (stats.disk[i].video+stats.disk[i].text);
    
    maxVideoBlocks = max(stats.disk[i].video, maxVideoBlocks);
    minVideoBlocks = min(stats.disk[i].video, minVideoBlocks);
    sumVideoBlocks += stats.disk[i].video;

    maxTextBlocks = max(stats.disk[i].text, maxTextBlocks);
    minTextBlocks = min(stats.disk[i].text, minTextBlocks);
    sumTextBlocks += stats.disk[i].text;
    
    /* reset per round vars to 0 */
    stats.disk[i].busyTime = 0 ;
    stats.disk[i].video = 0;
    stats.disk[i].text = 0;
  }

#ifdef STATS_DEBUG
 printf("stats : [clk = %f] max busy time = %f, max blocks = %f\n",clock,maxBusyTime, maxBlocks/diskArray.blockSize);
#endif
  record(maxBlocks, stats.maxBlocks);
  record(minBlocks, stats.minBlocks);
  record(sumBlocks/diskArray.numDisks, stats.sumBlocks);
  /*    record(maxVideoBlocks/diskArray.blockSize,stats.maxVideoBlocks);*/
  record(maxVideoBlocks,stats.maxVideoBlocks);
  record(minVideoBlocks, stats.minVideoBlocks);
  record(sumVideoBlocks/diskArray.numDisks, stats.sumVideoBlocks);
  record(maxTextBlocks, stats.maxTextBlocks);
  record(minTextBlocks,stats.minTextBlocks);
  record(sumTextBlocks/diskArray.numDisks,stats.sumTextBlocks);
  record(maxBusyTime,stats.maxBusyTime);
  record(minBusyTime,stats.minBusyTime);
  record(sumBusyTime/diskArray.numDisks,stats.sumBusyTime);

}


  
void
initStats(struct arguments &args, DiskArray &diskArray, Statistics &stats)
{
  int i,j;
  stats.disk = (Stats *) calloc(diskArray.numDisks, sizeof(Stats));
  if(null_ptr(stats.disk))
    error("calloc failed in initStats()\n",E_MEM);

  stats.pendingRequests = (PendingRequests **)calloc(args.numClients,sizeof(PendingRequests*));
  if(null_ptr(stats.pendingRequests))
    error("calloc failed in initStats()\n",E_MEM);

  for(i=0;i<args.numClients;i++){
    stats.pendingRequests[i] = (PendingRequests *) 
            calloc(MAX_PENDING_REQUESTS,sizeof(PendingRequests));
    if(null_ptr(stats.pendingRequests[i]))
      error("calloc failed in initStats()\n",E_MEM);
  }

  stats.maxBlocks = histogram("max blocks",50,0,1000);
  stats.minBlocks  = histogram("min blocks",50,0,1000);
  stats.sumBlocks  = histogram("sum blocks",75,0,5000);
  stats.maxBusyTime = histogram("max busy time",50,0,3*args.round);
  stats.minBusyTime = histogram("min busy time",50,0,3*args.round);
  stats.sumBusyTime = histogram("sum busy time",50,0,3*args.round);
  /*   stats.maxVideoBlocks = histogram("max video blocks",300,1,300);*/
     stats.maxVideoBlocks = histogram("max video blocks",50,1,1000);
  stats.minVideoBlocks = histogram("min video blocks",50,0,1000);
  stats.sumVideoBlocks = histogram("sum video blocks",50,0,1000);
  stats.maxTextBlocks = histogram("max text blocks",50,0,1000);
  stats.minTextBlocks = histogram("min text blocks",50,0,1000);
  stats.sumTextBlocks = histogram("sum text blocks",50,0,1000);
  stats.videoResponseTime = histogram("video response time",200,0,args.round);
  stats.textResponseTime = histogram("text response time",200,0,2*args.round);
  stats.textQueuingDelay = histogram("text queueing delay",200,0,2*args.round);
  stats.textServiceTime = histogram("text service time",200,0,2*args.round);
}

void
printReport(struct arguments &args, DiskArray &array, Statistics &stats)
{
 int i;
 printf("BusyTime max = %f, min = %f, avg = %f\n",table_mean(stats.maxBusyTime),
	table_mean(stats.minBusyTime), 
	table_mean(stats.sumBusyTime));
 printf("Blocks per disk per round max= %f, min = %f, avg = %f\n",table_mean(stats.maxBlocks)/diskArray.blockSize,
	table_mean(stats.minBlocks)/diskArray.blockSize, 
	table_mean(stats.sumBlocks)/diskArray.blockSize);
 /*   for(i=-1;i<histogram_num(stats.maxVideoBlocks)+1;i++)
     printf("[ %d  %d ] %d\n", i,  (i+1), histogram_bucket(stats.maxVideoBlocks,i+1)); 
     */
 if(args.serverPush == 0 ){
   printf("Video Response Time : max = %f, min = %f, avg = %f, var = %f\n",
	  table_max(stats.videoResponseTime),	  table_min(stats.videoResponseTime),	 
	  table_mean(stats.videoResponseTime),  table_var(stats.videoResponseTime));
   if(args.verboseOutput)
     for(i=-1;i<histogram_num(stats.videoResponseTime)+1;i++)
       printf("[ %f  %f ] %d\n", 5.0*i,  5.0*(i+1), histogram_bucket(stats.videoResponseTime,i+1));
 }
 if(args.numTextClients>0){
   printf("Text Response Time : max = %f, min = %f, avg = %f, var = %f queuing = %f, service = %f\n",
	  table_max(stats.textResponseTime),	  table_min(stats.textResponseTime),	 
	  table_mean(stats.textResponseTime),  table_var(stats.textResponseTime),
	  table_mean(stats.textQueuingDelay), table_mean(stats.textServiceTime));

   if(args.verboseOutput)
     for(i=-1;i<histogram_num(stats.textResponseTime)+1;i++)
       printf("[ %f  %f ] %d\n", 10.0*i, 10.0*(i+1), histogram_bucket(stats.textResponseTime,i+1));
 }
}

void
registerRequest(RequestList &req)
{
  int i,client,index;



  client = req.request[0].client;
  index= findEmptyPendingRequestSlot(client);
  if(index<0)
    error("registerRequest(): unable to find an empty slot in pendingRequests",E_GENERIC);
  stats.pendingRequests[client][index] = req.numRequests;
  for(i=0;i<req.numRequests;i++)
    req.request[i].pendingRequestIndex=index;
}

int
findEmptyPendingRequestSlot(int client)
{
  int i=0;
  int retval=-1;
  while(i<MAX_PENDING_REQUESTS){
    if(stats.pendingRequests[client][i]==0){
      retval=i;
      break;
    }
    i++;
  }
    return retval;
}

void
incrementPendingRequestSlot(Request &req)
{
  int i ;

  i = req.pendingRequestIndex;
    stats.pendingRequests[req.client][i]++;
}

void
decrementPendingRequestSlot(Request &req)
{
  int i ;

  i = req.pendingRequestIndex;
    stats.pendingRequests[req.client][i]--;

}

